`timescale 1ns / 1ps

module processing(REG_SELECT,
			DATA_IN,
			WE,
			CLK,
			GWE,
			RST,
			READY,
			READ_OUT,
			LEFT_OUT,
			RIGHT_OUT
		     );

	input  [3:0]  REG_SELECT;              	
	input  [15:0] DATA_IN;
   	input         WE, CLK, GWE, RST, READY;                  		
   	output [15:0] READ_OUT;			
	output [19:0] LEFT_OUT, RIGHT_OUT;		

	wire [15:0] READ_OUT;
	wire [19:0] LEFT_OUT, RIGHT_OUT;

	wire [15:0] sample_data, volume;
	wire [7:0]  read_addr, write_addr;
	wire        active_buffer, play;
	
  	wire        sample_data_we, volume_we, play_we, active_buffer_we, read_addr_we, write_addr_we; 

	assign sample_data_we    = (REG_SELECT[0] & WE);
  	assign volume_we         = (REG_SELECT[1] & WE);
	assign play_we           = (REG_SELECT[2] & WE);

	assign active_buffer_we  = ( ((read_addr == 8'hFF) & play) | ((write_addr == 8'hFF) & ~play) );
	assign read_addr_we      = play;
	assign write_addr_we     = sample_data_we;

 	assign READ_OUT = (REG_SELECT == 1) ? sample_data				:
                   	(REG_SELECT == 2) ? volume           			:
                   	(REG_SELECT == 4) ? { {{15}1'b0}, play } 			:
                   				  { {{15}1'b0}, active_buffer }     ;
			  

  	Nbit_reg #(16, 0)    sample_input(  .clk(CLK), 
                     	  	   		.we(sample_data_we), 
				  	   		.gwe(GWE),
                          	   		.rst(RST), 
                          	   		.in(DATA_IN), 
                          	   		.out(sample_data)
		       	       	   );

  	Nbit_reg #(16, 0)      volume_reg(  .clk(CLK), 
                     	  	   		.we(volume_we), 
				  	   		.gwe(GWE),
                          	   		.rst(RST), 
                          	   		.in(DATA_IN), 
                          	   		.out(volume)
		       	       	   );

  	Nbit_reg #(1, 0) 	       play_reg(  .clk(CLK), 
                     	  	   		.we(play_we), 
				  	   		.gwe(GWE),
                          	   		.rst(RST), 
                          	   		.in(DATA_IN[0]), 
                          	   		.out(play)
		       	       	   );

  	Nbit_reg #(1, 0)   active_buf_reg(  .clk(READY), 
                     	  	   		.we(active_buffer_we), 
				  	   		.gwe(1'b1),
                          	   		.rst(RST), 
                          	   		.in( (~active_buffer | ~play) ), 
                          	   		.out(active_buffer)
		       	       	   );

  	Nbit_reg #(8, 0)    read_addr_reg(  .clk(READY), 
                     	  	   		.we(read_addr_we), 
				  	   		.gwe(1'b1),
                          	   		.rst(RST), 
                          	   		.in( (read_addr + 1) ), 
                          	   		.out( read_addr )
		       	       	   );

  	Nbit_reg #(8, 0)   write_addr_reg(  .clk(CLK), 
                     	  	   		.we(write_addr_we), 
				  	   		.gwe(GWE),
                          	   		.rst(RST), 
                          	   		.in( (write_addr + 1) ), 
                          	   		.out( write_addr )
		       	       	   );

	//Storage of the audio data
	wire        buffer1_we, buffer2_we;
	wire [15:0] buffer1_out, buffer2_out;

	assign buffer1_we = (sample_data_we & ~(play ^ active_buffer));
	assign buffer2_we = (sample_data_we & (play ^ active_buffer));

	audio_buffer buffer1(read_addr, write_addr, READY, CLK, sample_data, buffer1_we, buffer1_out);
	audio_buffer buffer2(read_addr, write_addr, READY, CLK, sample_data, buffer2_we, buffer2_out);

	assign LEFT_OUT  = (active_buffer == 0) ? { {10{1'b0}}, buffer1_out[15:8], {2{1'b0}} } :
								{ {10{1'b0}}, buffer2_out[15:8], {2{1'b0}} } ;


	assign RIGHT_OUT = (active_buffer == 0) ? { {10{1'b0}}, buffer1_out[7:0], {2{1'b0}} } :
								{ {10{1'b0}}, buffer2_out[7:0], {2{1'b0}} } ;


endmodule