
module cf_read(	REG_SELECT,
			DATA_IN,
			READ_OUT,
			WE,
			CLK,
			REG_CLK,
			GWE,
			RST,
			MPA,
			MPD,
			MPCE_,
			MPWE_,
			MPOE_
		  ); 

  input [4:0] REG_SELECT;
  input [15:0] DATA_IN;
  input WE, CLK, REG_CLK, GWE, RST;
  output [15:0] READ_OUT;
  output [6:0] MPA;
  output MPCE_, MPWE_, MPOE_;
  inout [15:0] MPD;

  wire lba0_we, lba1_we, lba2_we, lba3_we, offset_we, ctrl_status_ext_we, ctrl_status_we;
  wire ctrl_status_gwe;  

  assign lba0_we = (REG_SELECT[0] & WE);
  assign lba1_we = (REG_SELECT[0] & WE);
  assign lba2_we = (REG_SELECT[1] & WE);
  assign lba3_we = (REG_SELECT[1] & WE);
  assign offset_we = ( (REG_SELECT[2] & WE) | (REG_SELECT[4] & ~WE) );
  assign ctrl_status_ext_we = (REG_SELECT[3] & WE);

  wire[3:0]  lba3_out;
  wire[7:0]  lba0_out, lba1_out, lba2_out, aceReadOut, cfDataHigh, cfDataLow;
  wire[15:0] offset_in, offset_out, ctrl_status_in, ctrl_status_out, cfData_in, cfRead_out;

  reg MPCE_, MPWE_, MPOE_;
  reg [6:0] MPA;

  reg ctrl_status_int_we;
  reg ctrl_status_sel;
  reg ctrl_status_rst;

  reg ace_mode;
  reg aceRead_we;
  reg aceReadCommit;
  reg aceWriteCommit;
  reg [1:0] readCounter;
  reg [1:0] writeCounter;
  reg [7:0] state;

  reg cfData_we;
  reg cfDataHigh_we;
  reg cfDataLow_we;

  reg [15:0] mpd_writeData;
  reg [7:0] numLoops;
  
  initial begin
    MPCE_ = 1;
    MPWE_ = 1;
    MPOE_ = 1;
    MPA = 0;

    ctrl_status_int_we = 0;
    ctrl_status_sel = 1;
	 ctrl_status_rst = 0;

    ace_mode = 0;
    aceRead_we = 0;
    aceReadCommit = 0;
    aceWriteCommit = 0;
    readCounter = 3;
    writeCounter = 2;
    state = 1;

    cfData_we = 0;
    cfDataHigh_we = 0;
    cfDataLow_we = 0;

    mpd_writeData = 0;
    numLoops = 256;
  end


  cf_bram cfData( .addr(offset_out),
		   .clk(CLK),
		   .din(cfData_in),
		   .dout(cfRead_out),
		   .we(cfData_we)
		 );

 assign READ_OUT = (REG_SELECT == 1) ? {lba1_out, lba0_out}       :
                   (REG_SELECT == 2) ? {4'h0, lba3_out, lba2_out} :
                   (REG_SELECT == 4) ? offset_out                 :
                   (REG_SELECT == 8) ? ctrl_status_out            :
			                    cfRead_out                 ;


  Nbit_reg #(8, 63) lba0( .clk(REG_CLK), 
                     	  .we(lba0_we), 
				  .gwe(GWE),
                          .rst(RST), 
                          .in(DATA_IN[7:0]), 
                          .out(lba0_out)
		       );

  Nbit_reg #(8, 0) lba1(  .clk(REG_CLK), 
                     	  .we(lba1_we), 
				  .gwe(GWE),
                          .rst(RST), 
                          .in(DATA_IN[15:8]), 
                          .out(lba1_out)
		       );

  Nbit_reg #(8, 0) lba2(  .clk(REG_CLK), 
                     	  .we(lba2_we), 
				  .gwe(GWE),
                          .rst(RST), 
                          .in(DATA_IN[7:0]), 
                          .out(lba2_out)
		       );

  Nbit_reg #(4, 0) lba3(  .clk(REG_CLK), 
                     	  .we(lba3_we), 
				  .gwe(GWE),
                          .rst(RST), 
                          .in(DATA_IN[11:8]), 
                          .out(lba3_out)
		       );

  mux_Nbit_2to1 #(16) offset_mux(	.a((offset_out + 1)),
						.b(DATA_IN),
						.sel(WE),
						.out(offset_in)
					  );


  Nbit_reg #(16, 0) offset(	.clk(REG_CLK), 
                     	    	.we(offset_we), 
				      .gwe(GWE),
                            	.rst(RST), 
                            	.in(offset_in), 
                            	.out(offset_out)
		         	  );

  or(ctrl_status_we, ctrl_status_int_we, ctrl_status_ext_we);
  or(ctrl_status_gwe, GWE, ctrl_status_int_we);
  mux_Nbit_2to1 #(16) ctrl_status_mux(	.a(16'h0000),
													.b(DATA_IN),
													.sel(ctrl_status_sel),
													.out(ctrl_status_in)
												 );

  Nbit_reg #(16, 16'h8000) ctrl_status(  	.clk(REG_CLK), 
														.we(ctrl_status_we),
				                              .gwe(ctrl_status_gwe), 
														.rst(RST | ctrl_status_rst), 
														.in(ctrl_status_in), 
														.out(ctrl_status_out)
												  );

  Nbit_reg #(8, 0) aceRead(   .clk(CLK), 
                     	      .we(aceRead_we),
										.gwe(1'b1),
                              .rst(RST), 
                              .in(MPD[7:0]), 
                              .out(aceReadOut)
		         );	


  Nbit_reg #(8, 0) cfHigh(    .clk(CLK), 
                     	      .we(cfDataHigh_we),
										.gwe(1'b1),
                              .rst(RST), 
                              .in(MPD[7:0]), 
                              .out(cfDataHigh)
		         );	

  Nbit_reg #(8, 0) cfLow(     .clk(CLK), 
                     	      .we(cfDataLow_we),
										.gwe(1'b1),
                              .rst(RST), 
                              .in(MPD[7:0]), 
                              .out(cfDataLow)
		         );	
  
  assign cfData_in = {cfDataHigh, cfDataLow}; 

  assign MPD = (ace_mode == 1) ? mpd_writeData [15:0] : 16'hz;
  

  always @(posedge CLK) begin
    if(ace_mode == 0)
    //READ SYSTEM_ACE
    begin
      case (readCounter)
        0: begin MPCE_ = 0; MPOE_ = 1; MPWE_ = 1; end
	     1: begin MPCE_ = 0; MPOE_ = 1; MPWE_ = 1; end
	     2: begin MPCE_ = 0; MPOE_ = 0; MPWE_ = 1; aceRead_we = 1; end
	     3: begin MPCE_ = 1; MPOE_ = 1; MPWE_ = 1; aceRead_we = 0; end
      endcase
	    
      if(readCounter == 3) begin
	if(aceReadCommit == 1)
	  readCounter = 0;
      end
      else
        readCounter = readCounter + 1;
    
    end	  
    else 
    //WRITE SYSTEM_ACE
    begin
      case (writeCounter)
        0: begin MPCE_ = 0; MPOE_ = 1; MPWE_ = 1; end
	     1: begin MPCE_ = 0; MPOE_ = 1; MPWE_ = 0; end
	     2: begin MPCE_ = 1; MPOE_ = 1; MPWE_ = 1; end
      endcase

      if(writeCounter == 2) begin
        if(aceWriteCommit == 1)
	  writeCounter = 0;
      end
      else
        writeCounter = writeCounter + 1;
      
    end
	  
  end



  reg [2:0] delay;
  always @(posedge CLK) begin
    cfDataHigh_we = (state == 29) & aceRead_we;
    cfDataLow_we = (state == 31) & aceRead_we;
    cfData_we = (state == 34);
	 
    //Write x02 to x18
    if(state == 2) begin
		ctrl_status_sel = 0;
		ctrl_status_int_we = 1;
		delay = 5;
      MPA = 7'h18;
		mpd_writeData = 16'h0002;
		ace_mode = 1;
		aceWriteCommit = 1;
    end
	
    //Read x04 
    if(state == 4) begin
		ctrl_status_int_we = 0;
    	MPA = 7'h4;
		ace_mode = 0;
		aceReadCommit = 1;
    end

    //Wait until bit 1 of x04 is set
    if(state == 6) begin
		if( (aceReadOut & 2) == 0 )
			state = 3;
    end
	 
    //Read x05
    if(state == 8) begin
      MPA = 7'h5;
		ace_mode = 0;
		aceReadCommit = 1;
    end
    
    //Wait until bit 0 of x05 is set
    if(state == 10) begin
		if( (aceReadOut & 1) == 0 )
			state = 7;
    end  

    //Write LBA[7:0] to x10
    if(state == 12) begin
		MPA = 7'h10;
		mpd_writeData[7:0] = lba0_out;
		ace_mode = 1;
		aceWriteCommit = 1;
    end
	
    //Write LBA[15:8]  to x11
    if(state == 14) begin
		MPA = 7'h11;
      mpd_writeData[7:0] = lba1_out;
		ace_mode = 1;
		aceWriteCommit = 1;
    end
	 
    //Write LBA[23:16] to x12
    if(state == 16) begin
      MPA = 7'h12;
      mpd_writeData[7:0] = lba2_out;
		ace_mode = 1;
		aceWriteCommit = 1;
    end
	
    //Write LBA[27:24] to x13
    if(state == 18) begin
      MPA = 7'h13;
      mpd_writeData[7:0] = lba3_out;
		ace_mode = 1;
		aceWriteCommit = 1;
    end
	
    //Write SECCNT = 1 to x14
    if(state == 20) begin
      MPA = 7'h14;
      mpd_writeData[7:0] = 8'h01;
		ace_mode = 1;
		aceWriteCommit = 1;
    end
	
    //Specify a read mem card data command (x3) to x15 
    if(state == 22) begin
    	MPA = 7'h15;
      mpd_writeData[7:0] = 8'h03;
		ace_mode = 1;
		aceWriteCommit = 1;
    end
	 
    //Read x04
    if(state == 24) begin
      MPA = 7'h4;
		ace_mode = 0;
		aceReadCommit = 1;
    end

    //Wait until bit 5 of x04 is set
    if(state == 26) begin
		if( (aceReadOut & 32) == 0 )
			state = 23;
    end
	
    //Read high-order CF data from x40 
    if(state == 28) begin
      MPA = 7'h40;
		ace_mode = 0;
		aceReadCommit = 1;
    end
	 
    //Read low-order CF data from x41
    if(state == 30) begin
      MPA = 7'h41;
		ace_mode = 0;
		aceReadCommit = 1;
    end

    //Add delay for odd states
    if((state % 2) != 0) begin
		aceReadCommit = 0;
		aceWriteCommit = 0;
		
		//In state 1, wait for a "begin" signal
		if((state == 1)) begin
 			ctrl_status_sel = 1;
         ctrl_status_int_we = 0;
			ctrl_status_rst = 0;
			if((ctrl_status_out != 0) & ~ctrl_status_out[15])
				state = state + 1;
		end
	
		else begin
			delay = delay - 1;
         if(delay == 0) begin
				delay = 5;
            state = state + 1;
			end
      end
		
    end
    
    //No delay for even states
    else
		state = state + 1;
	
	 //Decrement the loop iterator, and if the
	 //loop has completed, return to state 1
    if(state == 35) begin
		numLoops = numLoops - 1;
      
      if(numLoops == 0) begin
			ctrl_status_rst = 1;
			state = 1;
		end
      else
			state = 23;
    end

  end
  
endmodule


